/*
 * COPYRIGHT:        See COPYRIGHT.TXT
 * PROJECT:          Ext2 File System Driver for WinNT/2K/XP
 * FILE:             devctl.c
 * PROGRAMMER:       Matt Wu <mattwu@163.com>
 * HOMEPAGE:         http://ext2.yeah.net
 * UPDATE HISTORY: 
 */

/* INCLUDES *****************************************************************/

#include "ext2fs.h"

/* GLOBALS ***************************************************************/

extern PEXT2_GLOBAL Ext2Global;

/* DEFINITIONS *************************************************************/

NTSTATUS
Ext2DeviceControlCompletion (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PIRP             Irp,
    IN PVOID            Context);


#ifdef ALLOC_PRAGMA
#pragma alloc_text(PAGE, Ext2DeviceControl)
#pragma alloc_text(PAGE, Ext2DeviceControlNormal)
#pragma alloc_text(PAGE, Ext2ProcessVolumeProperty)
#pragma alloc_text(PAGE, Ext2ProcessUserProperty)
#pragma alloc_text(PAGE, Ext2RefreshVolumeMountPoint)
#pragma alloc_text(PAGE, Ext2QueryVolumeName)
#pragma alloc_text(PAGE, Ext2ProcessGlobalProperty)
#pragma alloc_text(PAGE, Ex2ProcessUserPerfStat)

#if EXT2_UNLOAD
#pragma alloc_text(PAGE, Ext2PrepareToUnload)
#endif
#endif


NTSTATUS
Ext2DeviceControlCompletion (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PIRP             Irp,
    IN PVOID            Context
    )
{
    if (Irp->PendingReturned) {
        IoMarkIrpPending(Irp);
    }

    return STATUS_SUCCESS;
}


NTSTATUS
Ext2DeviceControlNormal (IN PEXT2_IRP_CONTEXT IrpContext)
{
    PDEVICE_OBJECT  DeviceObject;
    BOOLEAN         CompleteRequest = TRUE;
    NTSTATUS        Status = STATUS_UNSUCCESSFUL;

    PEXT2_VCB       Vcb;

    PIRP            Irp;
    PIO_STACK_LOCATION IrpSp;
    PIO_STACK_LOCATION NextIrpSp;

    PDEVICE_OBJECT  TargetDeviceObject;
    
    __try {

        ASSERT(IrpContext != NULL);
        
        ASSERT((IrpContext->Identifier.Type == EXT2ICX) &&
            (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
        
        CompleteRequest = TRUE;

        DeviceObject = IrpContext->DeviceObject;
    
        if (IsExt2FsDevice(DeviceObject))  {
            Status = STATUS_INVALID_DEVICE_REQUEST;
            __leave;
        }
        
        Irp = IrpContext->Irp;
        IrpSp = IoGetCurrentIrpStackLocation(Irp);

        Vcb = (PEXT2_VCB) IrpSp->FileObject->FsContext;

        if (!((Vcb) && (Vcb->Identifier.Type == EXT2VCB) &&
              (Vcb->Identifier.Size == sizeof(EXT2_VCB)))) {
            Status = STATUS_INVALID_PARAMETER;
            __leave;
        }
        
        TargetDeviceObject = Vcb->TargetDeviceObject;
        
        //
        // Pass on the IOCTL to the driver below
        //
        
        CompleteRequest = FALSE;

        NextIrpSp = IoGetNextIrpStackLocation( Irp );
        *NextIrpSp = *IrpSp;

        IoSetCompletionRoutine(
            Irp,
            Ext2DeviceControlCompletion,
            NULL,
            FALSE,
            TRUE,
            TRUE );
        
        Status = IoCallDriver(TargetDeviceObject, Irp);

    } __finally  {

        if (!IrpContext->ExceptionInProgress) {
            if (IrpContext) {
                if (!CompleteRequest) {
                    IrpContext->Irp = NULL;
                }

                Ext2CompleteIrpContext(IrpContext, Status);
            }
        }
    }
    
    return Status;
}


#if EXT2_UNLOAD

NTSTATUS
Ext2PrepareToUnload (IN PEXT2_IRP_CONTEXT IrpContext)
{
    PDEVICE_OBJECT  DeviceObject;
    NTSTATUS        Status = STATUS_UNSUCCESSFUL;
    BOOLEAN         GlobalDataResourceAcquired = FALSE;
    
    __try {

        ASSERT(IrpContext != NULL);
        
        ASSERT((IrpContext->Identifier.Type == EXT2ICX) &&
            (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
        
        DeviceObject = IrpContext->DeviceObject;
        
        if (IsExt2FsDevice(DeviceObject)) {
            Status = STATUS_INVALID_DEVICE_REQUEST;
            __leave;
        }
        
        ExAcquireResourceExclusiveLite(
            &Ext2Global->Resource,
            TRUE );
        
        GlobalDataResourceAcquired = TRUE;
        
        if (FlagOn(Ext2Global->Flags, EXT2_UNLOAD_PENDING)) {
            DEBUG(DL_ERR, ( "Ext2PrepareUnload:  Already ready to unload.\n"));
            
            Status = STATUS_ACCESS_DENIED;
            
            __leave;
        }
        
        {
            PEXT2_VCB               Vcb;
            PLIST_ENTRY             ListEntry;

            ListEntry = Ext2Global->VcbList.Flink;

            while (ListEntry != &(Ext2Global->VcbList)) {

                Vcb = CONTAINING_RECORD(ListEntry, EXT2_VCB, Next);
                ListEntry = ListEntry->Flink;

                if (Vcb && (!Vcb->ReferenceCount) &&
                    IsFlagOn(Vcb->Flags, VCB_DISMOUNT_PENDING)) {
                    Ext2RemoveVcb(Vcb);
                    Ext2ClearVpbFlag(Vcb->Vpb, VPB_MOUNTED);

                    Ext2DestroyVcb(Vcb);
                }
            }
        }

        if (!IsListEmpty(&(Ext2Global->VcbList))) {

            DEBUG(DL_ERR, ( "Ext2PrepareUnload:  Mounted volumes exists.\n"));
            
            Status = STATUS_ACCESS_DENIED;
            
            __leave;
        }
        
        IoUnregisterFileSystem(Ext2Global->DiskdevObject);
        IoUnregisterFileSystem(Ext2Global->CdromdevObject);
        Ext2Global->DriverObject->DriverUnload = DriverUnload;
        SetLongFlag(Ext2Global->Flags ,EXT2_UNLOAD_PENDING);
        Status = STATUS_SUCCESS;

        DEBUG(DL_INF, ( "Ext2PrepareToUnload: Driver is ready to unload.\n"));

    } __finally {

        if (GlobalDataResourceAcquired) {
            ExReleaseResourceLite(&Ext2Global->Resource);
        }
        
        if (!IrpContext->ExceptionInProgress) {
            Ext2CompleteIrpContext(IrpContext, Status);
        }
    }
    
    return Status;
}

#endif

NTSTATUS
Ext2ProcessGlobalProperty(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PEXT2_VOLUME_PROPERTY2 Property
    )
{
    NTSTATUS        Status = STATUS_SUCCESS;
    BOOLEAN         GlobalDataResourceAcquired = FALSE;
    struct nls_table * PageTable = NULL;

    if (Property->Command != APP_CMD_SET_PROPERTY &&
        Property->Command != APP_CMD_SET_PROPERTY2 ) {
        Status = STATUS_INVALID_PARAMETER;
        goto errorout;
    }

    if (!IsFlagOn(Property->Flags, EXT2_FLAG_VP_SET_GLOBAL)) {
        Status = STATUS_INVALID_PARAMETER;
        goto errorout;
    }

    ExAcquireResourceExclusiveLite(&Ext2Global->Resource, TRUE);
    GlobalDataResourceAcquired = TRUE;

    if (Property->bReadonly) {
        ClearLongFlag(Ext2Global->Flags, EXT2_SUPPORT_WRITING);
        ClearLongFlag(Ext2Global->Flags, EXT3_FORCE_WRITING);
    } else {
        SetLongFlag(Ext2Global->Flags, EXT2_SUPPORT_WRITING);
        if (Property->bExt3Writable) {
            SetLongFlag(Ext2Global->Flags, EXT3_FORCE_WRITING);
        } else {
            ClearLongFlag(Ext2Global->Flags, EXT3_FORCE_WRITING);
        }
    }

    PageTable = load_nls(Property->Codepage);
    if (PageTable) {
        memcpy(Ext2Global->Codepage.AnsiName, Property->Codepage, CODEPAGE_MAXLEN);
        Ext2Global->Codepage.PageTable = PageTable;
    }

    if (Property->Command == APP_CMD_SET_PROPERTY2) {
        RtlZeroMemory(Ext2Global->sHidingPrefix, HIDINGPAT_LEN);
        if ((Ext2Global->bHidingPrefix = Property->bHidingPrefix)) {
            RtlCopyMemory( Ext2Global->sHidingPrefix,
                           Property->sHidingPrefix,
                           HIDINGPAT_LEN - 1);
        }
        RtlZeroMemory(Ext2Global->sHidingSuffix, HIDINGPAT_LEN);
        if ((Ext2Global->bHidingSuffix = Property->bHidingSuffix)) {
            RtlCopyMemory( Ext2Global->sHidingSuffix,
                           Property->sHidingSuffix,
                           HIDINGPAT_LEN - 1);
        }
    }

errorout:


    if (GlobalDataResourceAcquired) {
        ExReleaseResourceLite(&Ext2Global->Resource);
    }

    return Status;
}

NTSTATUS
Ext2QueryVolumeName(
    IN PVOID               DeviceObject,
    IN OUT PUNICODE_STRING DeviceName
    )
{

    NTSTATUS Status = STATUS_SUCCESS;
    ULONG    Length = 0;
    UNICODE_STRING  FileName;
    POBJECT_NAME_INFORMATION NameInfo = NULL;

    RtlZeroMemory(DeviceName, sizeof(UNICODE_STRING));

    /* querying length of volume device name */
    NameInfo = (POBJECT_NAME_INFORMATION) &FileName;
    RtlZeroMemory(&FileName, sizeof(UNICODE_STRING));
    Status = ObQueryNameString(
                      DeviceObject,
                      NameInfo,
                      sizeof(UNICODE_STRING),
                      &Length );

    if ( STATUS_INFO_LENGTH_MISMATCH == Status || 
         STATUS_BUFFER_OVERFLOW == Status ) {

        /* allocating memory for volume name */
        NameInfo = ExAllocatePoolWithTag(
                          NonPagedPool,
                          Length + 2,
                          EXT2_VNAME_MAGIC );

        if (NULL == NameInfo) {
            Status = STATUS_INSUFFICIENT_RESOURCES;
            goto errorout;
        }

        /* querying name of the volume device */
        Status = ObQueryNameString(
                          DeviceObject,
                          NameInfo,
                          Length + 2,
                          &Length );

        if (NT_SUCCESS(Status)) {

            /* moving name to the beginning of NameInfo */
            DeviceName->Buffer = (PWSTR)NameInfo;
            DeviceName->Length = NameInfo->Name.Length;
            DeviceName->MaximumLength = (USHORT)Length + 2;

            RtlMoveMemory ( DeviceName->Buffer,
                            &NameInfo->Name.Buffer[0],
                            DeviceName->Length );
            DeviceName->Buffer[DeviceName->Length/2] = 0;
        } else {

            ExFreePoolWithTag(NameInfo, EXT2_VNAME_MAGIC);
        }
    } else {
        if (Length <= sizeof(OBJECT_NAME_INFORMATION)) {
            Status = STATUS_UNSUCCESSFUL;
        }
    }

errorout:
    return Status;
}

NTSTATUS
Ext2RefreshVolumeMountPoint(
    IN PEXT2_VCB    Vcb,
    IN CHAR         DrvLetter
    )
{
    NTSTATUS    status = STATUS_SUCCESS;

    WCHAR   LinkBuffer[] = L"\\DosDevices\\Z:";
    UNICODE_STRING LinkName;

    UNICODE_STRING  VolumeName;

    RtlZeroMemory(&VolumeName, sizeof(UNICODE_STRING));

    if (DrvLetter >= '0' && DrvLetter <= '9') {
    } else if (DrvLetter >= 'A' && DrvLetter <= 'Z') {
    } else if (DrvLetter >= 'a' && DrvLetter <= 'z') {
        DrvLetter -= 0x20;
    } else {
        DrvLetter = 0;
    }
 
    if (!DrvLetter && !Vcb->DrvLetter) {
        goto errorout;
    }

    RtlInitUnicodeString(&LinkName, LinkBuffer);
    status = Ext2QueryVolumeName(
                    (PVOID)Vcb->RealDevice,
                    &VolumeName
                    );

    if (!NT_SUCCESS(status)) {
        goto errorout;
    }

    if (Vcb->DrvLetter) {
        LinkBuffer[12] = (USHORT)Vcb->DrvLetter;
        IoDeleteSymbolicLink(&LinkName);
    }

    if (DrvLetter) {
        Vcb->DrvLetter = DrvLetter;
        LinkBuffer[12] = (USHORT)DrvLetter;
        IoCreateSymbolicLink(&LinkName, &VolumeName);
    }

errorout:

    if (VolumeName.Buffer) {
        ExFreePoolWithTag(VolumeName.Buffer, EXT2_VNAME_MAGIC);
    }

    return status;
}

NTSTATUS
Ext2ProcessVolumeProperty(
    IN  PEXT2_VCB              Vcb,
    IN  PEXT2_VOLUME_PROPERTY2 Property
    )
{
    NTSTATUS        Status = STATUS_SUCCESS;
    BOOLEAN         VcbResourceAcquired = FALSE;
    struct nls_table * PageTable = NULL;

    ExAcquireResourceExclusiveLite(&Vcb->MainResource, TRUE);
    VcbResourceAcquired = TRUE;

    if (Property->Command == APP_CMD_SET_PROPERTY ||
        Property->Command == APP_CMD_SET_PROPERTY2 ) {

        if (Property->bReadonly) {
            SetLongFlag(Vcb->Flags, VCB_READ_ONLY);
        } else {
            if (!Vcb->IsExt3fs || (Vcb->IsExt3fs && Property->bExt3Writable))  {
                ClearLongFlag(Vcb->Flags, VCB_READ_ONLY);
            } else {
                SetLongFlag(Vcb->Flags, VCB_READ_ONLY);
            }
        }

        PageTable = load_nls(Property->Codepage);
        memcpy(Vcb->Codepage.AnsiName, Property->Codepage, CODEPAGE_MAXLEN);
        Vcb->Codepage.PageTable = PageTable;
        if (Vcb->Codepage.PageTable) {
            Ext2InitializeLabel(Vcb, Vcb->SuperBlock);
        }

        if (Property->Command == APP_CMD_SET_PROPERTY2) {

            Vcb->DrvLetter = Property->DrvLetter;
            /*
            the DosDevice Link to the volume deivce could not work :(
            Ext2RefreshVolumeMountPoint(Vcb, Property->DrvLetter);
            */

            RtlZeroMemory(Vcb->sHidingPrefix, HIDINGPAT_LEN);
            if (Vcb->bHidingPrefix = Property->bHidingPrefix) {
                RtlCopyMemory( Vcb->sHidingPrefix,
                               Property->sHidingPrefix,
                               HIDINGPAT_LEN - 1);
            }

            RtlZeroMemory(Vcb->sHidingSuffix, HIDINGPAT_LEN);
            if (Vcb->bHidingSuffix = Property->bHidingSuffix) {
                RtlCopyMemory( Vcb->sHidingSuffix,
                               Property->sHidingSuffix,
                               HIDINGPAT_LEN - 1);
            }
        }

    } else if (Property->Command == APP_CMD_QUERY_PROPERTY ||
               Property->Command == APP_CMD_QUERY_PROPERTY2 ) {

        Property->bExt2 = TRUE;
        Property->bExt3 = Vcb->IsExt3fs;
        Property->bReadonly = IsFlagOn(Vcb->Flags, VCB_READ_ONLY);
        if (!Property->bReadonly && Vcb->IsExt3fs) {
            Property->bExt3Writable = TRUE;
        } else {
            Property->bExt3Writable = FALSE;
        }

        RtlZeroMemory(Property->Codepage, CODEPAGE_MAXLEN);
        if (Vcb->Codepage.PageTable) {
            strcpy(Property->Codepage, Vcb->Codepage.PageTable->charset);
        } else {
            strcpy(Property->Codepage, "default");
        }

        if (Property->Command == APP_CMD_QUERY_PROPERTY2) {

            RtlCopyMemory(Property->UUID, Vcb->SuperBlock->s_uuid, 16);

            Property->DrvLetter = Vcb->DrvLetter;

            if (Property->bHidingPrefix = Vcb->bHidingPrefix) {
                RtlCopyMemory( Property->sHidingPrefix,
                               Vcb->sHidingPrefix,
                               HIDINGPAT_LEN);
            } else {
                RtlZeroMemory( Property->sHidingPrefix,
                               HIDINGPAT_LEN);
            }

            if (Property->bHidingSuffix = Vcb->bHidingSuffix) {
                RtlCopyMemory( Property->sHidingSuffix,
                               Vcb->sHidingSuffix,
                               HIDINGPAT_LEN);
            } else {
                RtlZeroMemory( Property->sHidingSuffix,
                               HIDINGPAT_LEN);
            }
        }

    } else {

        Status = STATUS_INVALID_PARAMETER;
        goto errorout;
    }


errorout:

    if (VcbResourceAcquired) {
        ExReleaseResourceLite(&Vcb->MainResource);
    }
       
    return Status;
}

NTSTATUS
Ext2ProcessUserProperty(
    IN PEXT2_IRP_CONTEXT IrpContext,
    IN  PEXT2_VOLUME_PROPERTY2 Property,
    IN  ULONG           Length
    )
{
    NTSTATUS    Status = STATUS_SUCCESS;
    PEXT2_VCB   Vcb = NULL;
    PDEVICE_OBJECT  DeviceObject = NULL;

    __try {

        ASSERT(IrpContext != NULL);
        
        ASSERT((IrpContext->Identifier.Type == EXT2ICX) &&
            (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));

        if (Property->Magic != EXT2_VOLUME_PROPERTY_MAGIC) {
            Status = STATUS_INVALID_PARAMETER;
            __leave;
        }

        DeviceObject = IrpContext->DeviceObject;
        if (IsExt2FsDevice(DeviceObject)) {
            Status = Ext2ProcessGlobalProperty(DeviceObject, Property);
        } else {
            Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
            if (!((Vcb) && (Vcb->Identifier.Type == EXT2VCB) &&
                           (Vcb->Identifier.Size == sizeof(EXT2_VCB)))) {
                Status = STATUS_INVALID_PARAMETER;
                __leave;
            }
            Status = Ext2ProcessVolumeProperty(Vcb, Property);
        }

        if (NT_SUCCESS(Status)) {
            IrpContext->Irp->IoStatus.Information = Length;
        }

    } __finally {

        if (!IrpContext->ExceptionInProgress) {
            Ext2CompleteIrpContext(IrpContext, Status);
        }
    }
    
    return Status;
}

NTSTATUS
Ex2ProcessUserPerfStat(
    IN PEXT2_IRP_CONTEXT IrpContext,
    IN  PEXT2_QUERY_PERFSTAT QueryPerf,
    IN  ULONG           Length
    )
{
    NTSTATUS    Status = STATUS_SUCCESS;
    PEXT2_VCB   Vcb = NULL;
    BOOLEAN     GlobalDataResourceAcquired = FALSE;

    PDEVICE_OBJECT  DeviceObject = NULL;

    __try {

        ASSERT(IrpContext != NULL);
        
        ASSERT((IrpContext->Identifier.Type == EXT2ICX) &&
            (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));

        DeviceObject = IrpContext->DeviceObject;
        if (IsExt2FsDevice(DeviceObject)) {

            if (QueryPerf->Magic != EXT2_QUERY_PERFSTAT_MAGIC) {
                Status = STATUS_INVALID_PARAMETER;
                __leave;
            }

            if (QueryPerf->Command != IOCTL_APP_QUERY_PERFSTAT) {
                Status = STATUS_INVALID_PARAMETER;
                __leave;
            }

            if (Length < sizeof(EXT2_QUERY_PERFSTAT)) {
                Status = STATUS_BUFFER_OVERFLOW;
                __leave;
            }

            ExAcquireResourceSharedLite(&Ext2Global->Resource, TRUE);
            GlobalDataResourceAcquired = TRUE;

            QueryPerf->PerfStat = Ext2Global->PerfStat;

        } else {
            Status = STATUS_INVALID_PARAMETER;
            __leave;
        }

        if (NT_SUCCESS(Status)) {
            IrpContext->Irp->IoStatus.Information = Length;
        }

    } __finally {

        if (GlobalDataResourceAcquired) {
            ExReleaseResourceLite(&Ext2Global->Resource);
        }

        if (!IrpContext->ExceptionInProgress) {
            Ext2CompleteIrpContext(IrpContext, Status);
        }
    }
    
    return Status;
}


NTSTATUS
Ext2DeviceControl (IN PEXT2_IRP_CONTEXT IrpContext)
{
    PIRP                Irp;
    PIO_STACK_LOCATION  irpSp;
    ULONG               code;
    ULONG               length;
    NTSTATUS            Status;
    
    ASSERT(IrpContext);
    
    ASSERT((IrpContext->Identifier.Type == EXT2ICX) &&
        (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
    
    Irp = IrpContext->Irp;
    
    irpSp = IoGetCurrentIrpStackLocation(Irp);
    
    code = irpSp->Parameters.DeviceIoControl.IoControlCode;
    length = irpSp->Parameters.DeviceIoControl.OutputBufferLength;
    
    switch (code) {

    case IOCTL_APP_VOLUME_PROPERTY:
        Status = Ext2ProcessUserProperty(
                        IrpContext,
                        Irp->AssociatedIrp.SystemBuffer,
                        length
                        );
        break;

    case IOCTL_APP_QUERY_PERFSTAT:
        Status = Ex2ProcessUserPerfStat(
                        IrpContext,
                        Irp->AssociatedIrp.SystemBuffer,
                        length
                        );
         break;

#if EXT2_UNLOAD
    case IOCTL_PREPARE_TO_UNLOAD:
        Status = Ext2PrepareToUnload(IrpContext);
        break;
#endif
    default:
        Status = Ext2DeviceControlNormal(IrpContext);
    }
    
    return Status;
}
